send_to :: ((Location, Light), Event) -> Rule ()
send_to ((location, light), e) = error "send_to"

luxOfLightScene :: Scene -> Lux
luxOfLightScene (S ls) = (sum . map f) ls
  where f (xs,l) = length xs * l * luxOfLight

malfunction_occurs :: Component -> Bool
malfunction_occurs(component) = 
 case component of
   OutdoorLightSensor(i)     -> not(fst(light_sensor(i)))
   MotionSensor(loc, i)      -> motion_detector(loc) == MDefect
   HallwayButton(hallway, i) -> hallway_button_defect(hallway)
   Luminaire(loc, light)     -> status_of_light(loc, light)
                                  /= status_line(loc, light)

lights_to_turn_on :: (Room, Lux, Scene) -> [(Light,DimValue)]
lights_to_turn_on (room, lux, scene@(S ls)) = 
   let neededLux       = max 0 (luxOfLightScene(scene)-lux)
       select 0 _      = []
       select i _
              | i <= 0 = []
       select _ []     = []
       select i ((lights,value) : ls)
              | lights `listIntersect` dlights /= []
                = let (failLights, okLights) = split (flip elem dlights) lights
                      (onLights, offLights)  = split f failLights
                      f l = light_defect(room,l) == Just Light_On
                      i'  = i - (length onLights) * maxDimValue * luxOfLight
                  in select i' ((okLights,value) : ls)
       select i ((lights,value) : ls)
              | i > length(lights) * value * luxOfLight
                = zip lights (repeat value) ++ 
                 select (i - length(lights) * value * luxOfLight) ls
              | otherwise
                = let v = i `div` (length(lights) * luxOfLight)
                  in zip lights (repeat v)
   in select neededLux ls
 where dlights = [ light | light <- lights_at(room)
                 , light_defect(room, light) /= Nothing
                 ]
 

status_line :: (Location, Light) -> LightStatus
status_line (location, light) = 
  case light_defect(location, light) of
   Nothing -> status_of_light(location, light)
   Just s  -> s

lightgroup_wall_button_pressed :: (Room, LightGroup) -> Bool
lightgroup_wall_button_pressed _ = True

hallway_button_pressed :: Hallway -> Bool
hallway_button_pressed (hallway) = True

any_hallway_button_defect :: Hallway -> Bool
any_hallway_button_defect (hallway) = hallway_button_defect(hallway)

switch_pressed :: (Room, PanelSwitch) -> Bool
switch_pressed = error "switch_pressed"

switch_value :: (Room, PanelSwitch) -> SwitchValue
switch_value = error "switch_value"

manually_switch_off_pressed :: Location -> Bool
manually_switch_off_pressed (loc) = True

no_event_for_ceiling_light :: Room -> Bool
no_event_for_ceiling_light (room) = True

switch                    = undefined
detect_unresonable_input  = undefined
report_energy_consumption = undefined
set_parameters            = undefined

somebody_is_moving :: Location -> Bool
somebody_is_moving (location) =
   case motion_detector(location) of
     MNoMotion  -> False
     MMotion(t) -> t >= current_time - fromInteger(motionCycle)
     MAlways    -> True
     MDefect    -> True


scene_all_on :: Room -> Scene
scene_all_on (room) = S [(lights_at(room), maxDimValue)]

available_scenes :: Room -> [(String, Scene)]
available_scenes (room) = [("default",       default_light_scene(room))
                          ,("light on desk", light_on_desk(room))
                          ,("all on",        scene_all_on(room))]


