all_locations :: [Location]
all_locations = all_rooms ++ all_hallways

all_rooms :: [Room]
all_rooms = [Room(i) | i <- [1..8]]

all_hallways :: [Hallway]
all_hallways = [Hallway(1)]

light_sensor :: AsmVar (Int -> (Bool, Lux))
light_sensor = initAssocs "light_sensor"
                    [ (1, (True, 0)), (2, (True, 0)) ]

lights_in_group :: (Room, LightGroup) -> [Light]
lights_in_group (Room(i), group) 
   | group == LightGroupWall   = [ Light(j)   | j <- [1..6]  ]
   | group == LightGroupWindow = [ Light(j)   | j <- [7..12] ]

lights_in_hallway :: Hallway -> [Light]
lights_in_hallway (Hallway(i)) = [Light(1)]


outdoor_light_sensor_defect :: Room -> Bool
outdoor_light_sensor_defect (Room(i))
  | i>=1 && i<5 = not(fst(light_sensor(1)))
  | otherwise   = not(fst(light_sensor(2)))

outdoor_sensor :: Room -> Lux
outdoor_sensor (Room(i))
  | i>=1 && i<5 = snd(light_sensor(1))
  | otherwise   = snd(light_sensor(2))

some_door_is_open :: Location -> Bool
some_door_is_open (location)
  | location_is_room(location)
    = door_is_open(location,1) || door_is_open(location,2) ||
      case location of
       Room(4) -> False
       Room(8) -> False
       Room(i) -> door_is_open(Room(i+1),2)
  | otherwise
    = door_is_open(location,1) || door_is_open(location,2) ||
      or [ door_is_open(r,1) | r <- all_rooms ]
    

door_is_open :: AsmVar ((Location, Int) -> Bool)
door_is_open = initAssocs "door_is_open" 
                    [ ((loc,j), False) | j <- [1,2], loc <- all_locations ]

rooms_under_lightsensor :: LightSensor -> [Room]
rooms_under_lightsensor (1) = [Room(i) | i <- [1..4] ]
rooms_under_lightsensor (2) = [Room(i) | i <- [5..8] ]


light_on_desk :: Room -> Scene
light_on_desk (room) = S 
   [([Light(1), Light(2), Light(3), Light(4)], maxDimValue)
   ,([Light(5)], 100)
   ,([Light(6), Light(7), Light(8)], 50)
   ,([Light(9), Light(10), Light(11), Light(12)], 80)
   ]

