type MWindow    = Maybe Window

type RegisterFunction = GUI () -> GUI ()

windowNotCreated :: Maybe Window
windowNotCreated = Nothing

guiMotion :: Motion -> Int
guiMotion MDefect   = 2
guiMotion MAlways   = 0
guiMotion _         = 1


onOffButtons :: String -> (Rule (), Rule ()) -> GUI () -> Window -> GUI Frame
onOffButtons name (ruleOn, ruleOff) updateGUI w = do
  l  <- label [text name] w
  on <- button [ text "on", command (do liftIO (fire1 ruleOn)
                                        updateGUI)
               , width 3] w
  off<- button [ text "off", command (do liftIO (fire1 ruleOff)
                                         updateGUI)
               , width 3] w
  f <- frame [] (l ^^ on ^^ off)
  result f

listboxWithTitle :: String -> Window -> GUI (Frame, Listbox [String])
listboxWithTitle title w = 
     do lb <- listbox [height 6, width 16] w
        v  <- vscroll [takeFocus False] lb
        l  <- label [text title] w
        f  <- frame [] (l ^-^ (flexible (flexible lb <|< v)))
        result (flexible f, lb)



createPanel :: Int -> GUI () -> GUI Window
createPanel room updateGUI = do
   w <- window [title ("panel for room " ++ show room)]
   swind <- onOffButtons "window" 
               ( switch_lightgroup_completely_on(Room(room), LightGroupWindow)
               , switch_lightgroup_off(Room(room), LightGroupWindow)
               ) updateGUI w
   swall <- onOffButtons "wall" 
               ( switch_lightgroup_completely_on(Room(room), LightGroupWall)
               , switch_lightgroup_off(Room(room), LightGroupWall)
               ) updateGUI w
   dummy <- label [] w
   ambient   <- button [text "ambient"] w
   (f,scene) <- listboxWithTitle "Light Scenes" w
   setValue scene (map fst (available_scenes(Room(room))))
   cset ambient (command (do 
          (liftIO(fire1(activate_light_scene(Room(room)
                                            ,last_light_scene(Room(room))))))
          updateGUI))
   cset scene (on (click 1) (do
          [i] <- getSelection scene
          liftIO (fire1 (set_light_scene(Room(room),
                          snd(available_scenes(Room(room))!!i))))
          done))
   pack ( (ambient << swind << swall)
          ^^ fillX dummy
          ^^ flexible f)
   result w

showPanel :: Int -> GUI () -> GVar MWindow -> GUI ()
showPanel room updateGUI guiPanel = do
   panel <- readGVar guiPanel
   case panel of
    Nothing -> do w <- createPanel room updateGUI
                  writeGVar guiPanel (Just w)
    Just w  -> do s <- tk_getTcl ["wm","state", tk_getPathName w]
                  if s == "iconic"
                   then tk_putTcl ["wm","deiconify", tk_getPathName w]
                   else tk_putTcl ["wm","iconify", tk_getPathName w]

topRoomWall :: GUI () -> Int -> Window -> GUI Frame
topRoomWall updateGUI room w = do
  door  <- button [text "closed", relief "flat", width 5] w
  swind <- button [text "window"] w
  swall <- button [text "wall"]   w
  dummy <- button [relief "flat"] w
  panel <- button [text "panel"] w
  f     <- frame  [] (panel << flexible dummy << door << swind << swall)
  color <- cget door background
  closeDoor door color 1
  guiPanel <- newGVar windowNotCreated
  cset panel (command (showPanel room updateGUI guiPanel))
  cset swind (command (do
               liftIO $ fire1
                  (room_wall_button(Room(room), LightGroupWindow))
               updateGUI))
  cset swall (command (do
               liftIO $ fire1
                  (room_wall_button(Room(room), LightGroupWall))
               updateGUI))
  result (fillX f)
 where openDoor d c j = do 
             csets d [ text "open", background "blue"
                     , command (closeDoor d c j) ]
             liftIO (fire1 (door_is_open(Room(room),j) := True))
       closeDoor d c j = do
             csets d [ text "closed", background c
                     , command (openDoor d c j) ]
             liftIO (fire1 (door_is_open(Room(room),j) := False))

leftRoomWall :: GUI () -> Int -> Window -> GUI Frame
leftRoomWall updateGUI i w = do
  dummy1 <- button [relief "flat"] w
  dummy2 <- button [relief "flat"] w
  door   <- button [] w
  color  <- cget door background
  closeDoor door color 2
  if (i==1 || i==5)
   then result (fillY (dummy1 ^*|^ dummy2))
   else result (fillY (dummy1 ^*|^ door ^*|^ dummy2))
 where openDoor d c j = do 
             csets d [ background "blue"
                     , command (closeDoor d c j) ]
             liftIO (fire1 (door_is_open(Room(i),j) := True))
       closeDoor d c j = do
             csets d [ background c
                     , command (openDoor d c j) ]
             liftIO (fire1 (door_is_open(Room(i),j) := False))

rightRoomWall :: Window -> GUI Frame
rightRoomWall w = do
  dummy <- button [relief "flat"] w
  f <- frame [] (flexible dummy)
  result (fillY f)

bottomRoomWall :: Int -> Window -> GUI Frame
bottomRoomWall i w = do
  always <- radiobutton [text "A"] w
  manual <- radiobutton [text "M"] w
  defect <- radiobutton [text "D"] w
  cset manual (command (
           liftIO $ fire1 (motion_detector(Room(i)) := MMotion(current_time))))
  cset always (command (
           liftIO $ fire1 (motion_detector(Room(i)) := MAlways)))
  cset defect (command (
           liftIO $ fire1 (motion_detector(Room(i)) := MDefect)))
  radio [ initValue (guiMotion (motion_detector(Room(i))))] 
        [always, manual, defect]
  f <- frame [] (flexible (always << manual << defect))
  result (fillX f)

cLight :: GUI () -> (Location, Light) -> GUI ()
cLight update (location, light) = do
  liftIO (fire1 (light_defect(location, light) :=
             case light_defect(location, light) of
               Nothing        -> Just Light_On
               Just Light_On  -> Just Light_Off
               Just Light_Off -> Nothing))
  update

ceilingLights :: GUI () -> Window -> Room -> [Light] -> GUI (GUI (), Frame)
ceilingLights update w room lights = do
  lights <- binds [ do left  <- canvas [background "black", 
                                        width 5, height 5] w
                       right <- canvas [background "black", 
                                        width 5, height 5] w
                       cset left  (on (click 1) (cLight update (room,li)))
                       cset right (on (click 1) (cLight update (room,lj)))
                       result (((left, li) , 
                                (right,lj)), flexible (left <*+< right))
                  | [li, lj] <- groupn 2 lights
                  ]
  let f = vertical (map snd lights)
  let all = concat . map (\((a,b),_) -> [a,b]) $ lights
  let act = let ss light  = case light_defect(room, light) of
                              Nothing        -> "#" ++ s ++ s ++ s
                              Just Light_On  -> "green"
                              Just Light_Off -> "red"
                  where s     = int2hex value
                        value = dim_value(room, light)
            in seqs [ cset l (background (ss light)) | (l,light) <- all ]
  dynamicExpr act
  result (act, flexible f)


createRoom :: RegisterFunction -> GUI () -> Int -> String -> GUI Window
createRoom regFun updateGUI i pos = do
   w <- window [title ("Room " ++ show i)]
   topWall    <- topRoomWall updateGUI i w
   leftWall   <- leftRoomWall updateGUI i w
   rightWall  <- rightRoomWall w
   bottomWall <- bottomRoomWall i w
   (act1, walllights) <- ceilingLights updateGUI w (Room(i))
            (let lights = lights_in_group(Room(i), LightGroupWall)
             in if pos=="t" then lights else reverse(lights))
   bar <- label [] w
   (act2, windlights) <- ceilingLights updateGUI w (Room(i))
            (let lights = lights_in_group(Room(i), LightGroupWindow)
             in if pos=="t" then reverse(lights) else lights)
   if pos == "t" then
     pack (leftWall << flexible (topWall 
                                 ^^ flexible (flexible walllights ^^ 
                                             (fillY bar ^^ 
                                              flexible windlights))
                                 ^^ bottomWall)
                    << rightWall)
    else
     pack (leftWall << flexible (bottomWall 
                                 ^^ flexible (flexible windlights ^^
                                             (fillY bar ^^
                                              flexible walllights))
                                 ^^ topWall)
                    << rightWall)
   regFun (dynamicExpr (do act1; act2))
   result w

createMalfunctionRoom :: Int -> GUI (Window, GUI ())
createMalfunctionRoom i = do
   w <- window [title ("malfunctions for room " ++ show i)]
   b <- button [text "update"] w
   (f,lb) <- listboxWithTitle "malfunctions" w
   pack (flexible (flexible f ^^ fillX b))
   let act = dynamicExpr $ do
              let ls = [(t,m) | ((r,t,m),_) <- assocs(user_information)
                       , r == Room(i)
                       ]
              setValue lb (map show ls)
              tk_toTcl lb ["see","end"]
   cset b (command act)
   act
   result (w, act)

createDatabase :: GUI (Window, GUI ())
createDatabase = do
   w <- window [title "malfunction database"]
   b <- button [text "update"] w
   (f,lb) <- listboxWithTitle "database" w
   pack (flexible (flexible f ^^ fillX b))
   let act = dynamicExpr $ do
              let ls = map fst (assocs(database))
              setValue lb (map show ls)
              tk_toTcl lb ["see","end"]
   cset b (command act)
   act
   result (w, act)

createHallway :: RegisterFunction -> GUI () -> Int -> GUI Window
createHallway regFun updateGUI i = do
   w <- window [title "Hallway"]
   switch <- button [text "switch"] w
   always <- radiobutton [text "A"] w
   manual <- radiobutton [text "M"] w
   defect <- radiobutton [text "D"] w
   cset manual (command (
      liftIO $ fire1 (motion_detector(Hallway(i)) := MMotion(current_time))))
   cset always (command (
      liftIO $ fire1 (motion_detector(Hallway(i)) := MAlways)))
   cset defect (command (
      liftIO $ fire1 (motion_detector(Hallway(i)) := MDefect)))
   radio [initValue (guiMotion (motion_detector(Hallway(i))))] 
         [always, manual, defect]
   f <- frame [] (flexible (switch << (always << manual << defect)))
   cset switch (command (do
          liftIO (fire1(hallway_button(Hallway(i))))
          updateGUI))
   left  <- button [] w
   right <- button [] w
   color <- cget left background
   closeDoor left  color 1
   closeDoor right color 2
   light <- canvas [background "black", width 5, height 1] w 
   cset light (on (click 1) (cLight updateGUI (Hallway(1),Light(1))))
   pack (flexible (fillY left << flexible light << fillY right)
         ^^ f)
   let act = dynamicExpr $
             let ss = case light_defect(Hallway(i), Light(1)) of
                        Nothing -> if status_of_light(Hallway(i), Light(1)) 
                                         == Light_Off
                                   then "black"
                                   else "white"
                        Just Light_On  -> "green"
                        Just Light_Off -> "red"
             in cset light (background ss)
   regFun act
   act
   result w
 where openDoor d c j = do 
             csets d [  background "blue"
                     , command (closeDoor d c j) ]
             liftIO (fire1 (door_is_open(Hallway(i),j) := True))
       closeDoor d c j = do
             csets d [ background c
                     , command (openDoor d c j) ]
             liftIO (fire1 (door_is_open(Hallway(i),j) := False))

createView :: RegisterFunction -> RegisterFunction 
    -> GUI () -> GUI ()
createView regFun regMFun updateGUI = do
  w <- window [title "Floor"]
  top    <- binds [ createR w i "b" | i <- [1..4] ]
  bottom <- binds [ createR w i "t" | i <- [5..8] ]
  hallway <- createH w 1
  ltb <- checkbutton [text "Ok", initValue True] w
  lt <- hscale [ scaleRange (0,10000)
               , text "light sensor in lux"
               , height 200
               , initValue 0
               ] w
  lbb <- checkbutton [text "Ok", initValue True] w
  lb <- hscale [ scaleRange (0,10000)
               , text "light sensor in lux"
               , height 200
               , initValue 0
               ] w
  let updateSensor1 = do v <- getValue lt
                         b <- getValue ltb
                         liftIO (fire1(light_sensor(1) := (b,v)))
  let updateSensor2 = do v <- getValue lb
                         b <- getValue lbb
                         liftIO (fire1(light_sensor(2) := (b,v)))
  cset lt  (command updateSensor1)
  cset ltb (command updateSensor1)
  cset lb  (command updateSensor2)
  cset lbb (command updateSensor2)
  ftop    <- frame [borderWidth 3, background "black"] 
                   (horizontal (map flexible top))
  fbottom <- frame [borderWidth 3, background "black"] 
                   (horizontal (map flexible bottom))
  pack $ ((fillY ltb << lt)
          ^*+^ ftop
          ^*+^ hallway 
          ^*+^ fbottom
          ^*+^ (fillY lbb << lb))
 where
   conf_frame = [borderWidth 2, relief "raised"]
   createR w i pos = do
       b  <- button [text (show i), relief "sunken"] w
       bm <- button [text "M", relief "sunken"] w
       color <- cget b background
       bOff <- button [text "Off"] w
       room <- newGVar windowNotCreated
       malfunction <- newGVar windowNotCreated
       cset b  (command (openRoom i room pos))
       cset bm (command (openMalfunction i malfunction))
       cset bOff (command (do liftIO (fire1(manually_switch_off(Room(i))))
                              updateGUI))
       let act = dynamicExpr $
                 if occupied(Room(i))
                 then cset b (background "yellow")
                 else cset b (background color)
       regFun act
       f <- frame conf_frame (flexible (b <*+< bm) ^^ fillX bOff)
       result f
   createH w i = do
       b       <- button [text "Hallway", relief "sunken"] w
       color   <- cget b background
       defect  <- checkbutton [text "Ok", initValue True] w
       cset defect (command (do
             v <- getValue defect
             liftIO (fire1(hallway_button_defect(Hallway(i)) := not(v)))))
       bOff    <- button [text "Off"] w
       hallway <- newGVar windowNotCreated
       cset b (command (openHallway hallway i))
       cset bOff (command (do liftIO (fire1(manually_switch_off(Hallway(1))))
                              updateGUI))
       let act = dynamicExpr $
                 if occupied(Hallway(i))
                 then cset b (background "yellow")
                 else cset b (background color)
       regFun act
       f <- frame (conf_frame ++ [borderWidth 3, background "black"])
                  (flexible b ^^ 
                   flexible (fillY defect << expand (fillX bOff)))
       result f
   openMalfunction i malfunction = do 
       r <- readGVar malfunction
       case r of
        Nothing -> do (w, act) <- createMalfunctionRoom i
                      writeGVar malfunction (Just w)
                      regMFun act
        Just w  -> do s <- tk_getTcl ["wm","state", tk_getPathName w]
                      if s == "iconic"
                       then tk_putTcl ["wm","deiconify", tk_getPathName w]
                       else tk_putTcl ["wm","iconify", tk_getPathName w]
   openRoom i room pos = do 
       r <- readGVar room
       case r of
        Nothing -> do w <- createRoom regFun updateGUI i pos
                      writeGVar room (Just w)
        Just w  -> do s <- tk_getTcl ["wm","state", tk_getPathName w]
                      if s == "iconic"
                       then tk_putTcl ["wm","deiconify", tk_getPathName w]
                       else tk_putTcl ["wm","iconify", tk_getPathName w]
   openHallway hallway i = do
       h <- readGVar hallway
       case h of
        Nothing -> do w <- createHallway regFun updateGUI i
                      writeGVar hallway (Just w)
        Just w  -> do s <- tk_getTcl ["wm","state", tk_getPathName w]
                      if s == "iconic"
                       then tk_putTcl ["wm","deiconify", tk_getPathName w]
                       else tk_putTcl ["wm","iconify", tk_getPathName w]

maxEnergie :: Float
maxEnergie = fromInteger((1+length(all_rooms)*12) * maxDimValue)

createControl :: GUI (RegisterFunction, RegisterFunction, GUI (), GUI ())
createControl = do
  w <- window [title "control"]
  let dflt = 400
  t  <- timer [initValue dflt, active False]
  ts <- hscale [ scaleRange (1,3000)
               , text "delay in ms"
               , height 160
               , initValue dflt
               ] w
  ms <- hscale [ scaleRange (1,10)
               , text "motion-cycle"
               , height 80
               , initValue motionCycle
               ] w
  ls <- hscale [ scaleRange (1,10)
               , text "light-cycle"
               , height 80
               , initValue lightCycle
               ] w
  cset ts (command (do v <- getValue ts
                       setValue t v))
  cset ms (command (do v <- getValue ms
                       liftIO (fire1(motionCycle := v))))
  cset ls (command (do v <- getValue ls
                       liftIO (fire1(lightCycle := v))))
  bc <- button [text "quit", width 5] w
  bs <- button [text "start", width 5] w
  let stop  = do cset t  (active False)
                 cset bs (text "start")
                 cset bs (command start)
      start = do cset bs (text "stop")
                 cset bs (command stop)
                 cset t  (active True)
  cset bs (command start)
  cset bc (command (do cset t (active False)
                       quit))
  actions  <- newGVar ([] :: [GUI ()])
  mactions <- newGVar ([] :: [GUI ()])
  lenergie <- label [text "Energy"] w
  energie  <- canvas [background "white", width 80, height 5] w
  crect (0,0) (0,0) [fillColor "red"] energie
  let update    = updateGUI actions updateEnergie
      updateM   = updateGUI mactions done
      updateEnergie = do
                  w <- tk_getTcl ["winfo","width",tk_getPathName energie]
                       `bind` tk_fromGUI
                  h <- tk_getTcl ["winfo","height",tk_getPathName energie]
                       `bind` tk_fromGUI
                  if (w /= maxHistory) 
                   then liftIO (fire1(maxHistory := w))
                   else done
                  clearCanvas energie
                  let f x = fromInteger(h) * (fromInteger(x) / maxEnergie)
                      ps  = zipWith (\e x -> (x,h-truncate(f e)))
                                    (0 : reverse (take (w-2) energieH)) [0..]
                            ++ [(length energieH - 1,h)]
                  cpolygon ps [penColor "black"] energie
      actTime   = dynamicExpr $ do
                    putStrLn (show current_time)
                    liftIO (do
                      fire1 (do current_time := current_time + 1.0
                                iUpdateMalfunction := iUpdateMalfunction + 1
                                iMotionCycle := iMotionCycle + 1
                                iLightCycle  := iLightCycle + 1)
                      if iMotionCycle>=motionCycle 
                       then fire1 (do iMotionCycle := 0
                                      motion_rules)
                       else done
                      if iLightCycle>=lightCycle 
                        then fire1 (do iLightCycle := 0
                                       malfunction
                                       light_rules)
                        else done)
                    if iUpdateMalfunction>=10
                       then do (liftIO (fire1 (iUpdateMalfunction := 0)))
                               updateM
                       else done
                    if iMotionCycle==0 || iLightCycle == 0
                     then update
                     else done
  cset t (command actTime)
  pack ((flexible (flexible ts ^^ 
                   flexible (flexible ms << flexible ls) ^^ bc << bs))
        <*+< (lenergie ^^ flexible energie))
  result (addAction actions, addAction mactions, update, updateM)
 where addAction :: GVar [GUI ()] -> GUI () -> GUI ()
       addAction actions act = do
                    acts <- readGVar actions
                    writeGVar actions (act : acts)
       updateGUI actions updateEnergie = do 
                    acts <- readGVar actions
                    updateEnergie
                    seqs acts
                    updateTask

iMotionCycle :: AsmVar Int
iMotionCycle = initVal "iMotionCycle" 0

iLightCycle :: AsmVar Int
iLightCycle = initVal "iLightCycle" 0

iUpdateMalfunction :: AsmVar Int
iUpdateMalfunction = initVal "iUpdateMalfunction" 0


int2hex :: Int -> String
int2hex i = [ c(i `div` 16), c(i `mod` 16) ]
 where c i | i <10     = chr (ord '0' + i)
           | otherwise = chr (ord 'a' + i - 10)

value :: AsmVar Int
value = initVal "value" 0


startGUI :: GUI ()
startGUI = do
   (regFun, regMFun, updateGUI, mupdateGUI) <- createControl
   createView regFun regMFun updateGUI
   (w,act) <- createDatabase
   regMFun act
   

main :: IO ()
main = start $ startGUI

l :: [AsmVar Int]
l = [ f "l" i | i <- [1..10] ]
  where f = initVal

g = l!!0 + 1

