data Ordering = LT | EQ | GT


compare :: Ord a => a -> a -> Ordering
compare x y | x > y     = GT
            | x < y     = LT
            | otherwise = EQ

sortBy          :: (a -> a -> Ordering) -> [a] -> [a]
sortBy cmp       = foldr (insertBy cmp) []

insertBy        :: (a -> a -> Ordering) -> a -> [a] -> [a]
insertBy cmp x []    = [x]
insertBy cmp x ys@(y:ys')
             = case cmp x y of
                GT -> y : insertBy cmp x ys'
                _  -> x : ys

listIntersect               :: (Eq a) => [a] -> [a] -> [a]
listIntersect                = listIntersectBy (==)

listDifference :: Eq a => [a] -> [a] -> [a]
listDifference = (\\)

listIntersectBy             :: (a -> a -> Bool) -> [a] -> [a] -> [a]
listIntersectBy eq xs ys     = [x | x <- xs, any (eq x) ys]

groupBy                 :: (a -> a -> Bool) -> [a] -> [[a]]
groupBy eq []            = []
groupBy eq (x:xs)        = (x:ys) : groupBy eq zs
                           where (ys,zs) = span (eq x) xs

head' s [] = error s
head' s xs = head xs

split :: (a -> Bool) -> [a] -> ([a], [a])
split p []     = ([], [])
split p (x:xs) = let (as,bs) = split p xs
                 in if p x then (x:as, bs) else (as,x:bs)
